# Copyright 2006-2009 Scott Horowitz <stonecrest@gmail.com>
# Copyright 2009-2014 Jonathan Ballet <jon@multani.info>
#
# This file is part of Sonata.
#
# Sonata is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Sonata is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Sonata.  If not, see <http://www.gnu.org/licenses/>.

import configparser
from io import StringIO
import logging
import os
import pkgutil
import re

import sonata.plugins

def find_plugin_dirs():
    return [
        os.path.expanduser('~/.config/sonata/plugins'),
        '/usr/local/lib/sonata/plugins']

# add dirs from sys.path:
sonata.plugins.__path__ = pkgutil.extend_path(sonata.plugins.__path__,
                          sonata.plugins.__name__)
# add dirs specific to sonata:
sonata.plugins.__path__ = find_plugin_dirs() + sonata.plugins.__path__


class Plugin:
    def __init__(self, path, name, info, load):
        self.logger = logging.getLogger(__name__)
        self.path = path
        self.name = name
        self._info = info
        self._load = load
        # obligatory plugin info:
        format_value = info.get('plugin', 'plugin_format')
        self.plugin_format = tuple(map(int, format_value.split(',')))
        self.longname = info.get('plugin', 'name')
        versionvalue = info.get('plugin', 'version')
        self.version = tuple(map(int, versionvalue.split(',')))
        self.version_string = '.'.join(map(str, self.version))
        self._capabilities =  dict(info.items('capabilities'))
        # optional plugin info:
        try:
            self.description = info.get('plugin', 'description')
        except configparser.NoOptionError:
            self.description = ""
        try:
            self.author = info.get('plugin', 'author')
        except:
            self.author = ""
        try:
            self.author_email = info.get('plugin', 'author_email')
        except:
            self.author_email = ""
        try:
            self.iconurl = info.get('plugin', 'icon')
        except configparser.NoOptionError:
            self.iconurl = None
        try:
            self.url = info.get('plugin', 'url')
        except:
            self.url = ""
        # state:
        self._module = None # lazy loading
        self._enabled = False

    def get_enabled(self):
        return self._enabled

    def get_features(self, capability):
        if not self._enabled or not capability in self._capabilities:
            return []

        module = self._get_module()
        if not module:
            return []

        features = self._capabilities[capability]
        try:
            return [self.get_feature(module, f)
                for f in features.split(', ')]
        except KeyboardInterrupt:
            raise
        except:
            self.logger.exception("Failed to access features in plugin %s.",
                                  self.name)
            return []

    def get_feature(self, module, feature):
        obj = module
        for name in feature.split('.'):
            obj = getattr(obj, name)
        return obj

    def _get_module(self):
        if not self._module:
            try:
                self._module = self._load()
            except Exception:
                self.logger.exception("Failed to load plugin %s.", self.name)
                return None
        return self._module

    def force_loaded(self):
        return bool(self._get_module())


class BuiltinPlugin(Plugin):
    def __init__(self, name, longname, description, capabilities, object):
        self.name = name
        self.longname = longname
        self.description = description
        self._capabilities = capabilities
        self._module = object
        self.version_string = "Built-in"
        self.author = self.author_email = self.url = ""
        self.iconurl = None
        self._enabled = False

    def _get_module(self):
        return self._module


class PluginSystem:
    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self.plugin_infos = []
        self.notifys = []

    def get_info(self):
        return self.plugin_infos

    def get(self, capability):
        return [(plugin, feature)
            for plugin in self.plugin_infos
            for feature in plugin.get_features(capability)]

    def get_from_name(self, name):
        for plugin in self.plugin_infos:
            if plugin.longname == name:
                return plugin
        return None

    def notify_of(self, capability, enable_cb, disable_cb):
        self.notifys.append((capability, enable_cb, disable_cb))
        for plugin, feature in self.get(capability):
            enable_cb(plugin, feature)

    def set_enabled(self, plugin, state):
        if plugin._enabled != state:
            # make notify callbacks for each feature of the plugin:
            plugin._enabled = True # XXX for plugin.get_features

            # process the notifys in the order they were registered:
            order = (lambda x:x) if state else reversed
            for capability, enable, disable in order(self.notifys):
                callback = enable if state else disable
                for feature in plugin.get_features(capability):
                    callback(plugin, feature)
            plugin._enabled = state

    def find_plugins(self):
        for path in sonata.plugins.__path__:
            if not os.path.isdir(path):
                continue
            for entry in os.listdir(path):
                if entry.startswith('_'):
                    continue # __init__.py etc.
                if entry.endswith('.py'):
                    try:
                        self.load_info(path, entry[:-3])
                    except:
                        self.logger.exception("Failed to load info: %s",
                                              os.path.join(path, entry))

    def load_info(self, path, name):
        f = open(os.path.join(path, name+".py"), "rt", encoding="utf-8")
        text = f.read()
        f.close()

        pat = r'^### BEGIN PLUGIN INFO.*((\n#.*)*)\n### END PLUGIN INFO'
        infotext = re.search(pat, text, re.MULTILINE).group(1)
        uncommented = '\n'.join(line[1:].strip()
                    for line in infotext.split('\n'))
        info = configparser.ConfigParser()
        info.read_file(StringIO(uncommented))

        plugin = Plugin(path, name, info,
                lambda:self.import_plugin(name))

        # add only newest version of each name
        old_plugin = self.get_from_name(plugin.longname)
        if old_plugin:
            if plugin.version > old_plugin.version:
                self.plugin_infos.remove(old_plugin)
                self.plugin_infos.append(plugin)
        else:
            self.plugin_infos.append(plugin)

        if not info.options('capabilities'):
            self.logger.warning("No capabilities in plugin %s.", name)

    def import_plugin(self, name):
        # XXX load from a .py file - no .pyc etc.
        __import__('sonata.plugins', {}, {}, [name], 0)
        plugin = getattr(sonata.plugins, name)
        return plugin

pluginsystem = PluginSystem()




# this is the magic interpreted by Sonata, referring to tab_construct below:

### BEGIN PLUGIN INFO
# [plugin]
# plugin_format: 0, 0
# name: Local MPD
# version: 0, 0, 1
# description: A tab for controlling local MPD.
# author: Tuukka Hastrup
# author_email: Tuukka.Hastrup@iki.fi
# url: http://www.nongnu.org/sonata/
# license: GPL v3 or later
# [capabilities]
# tab_construct: tab_construct
### END PLUGIN INFO

import subprocess, locale, os
from pwd import getpwuid

from gi.repository import GLib, Gtk

from sonata.misc import escape_html
from sonata import ui


class Netstat:
    TCP_STATE_NAMES = ("ESTABLISHED SYN_SENT SYN_RECV FIN_WAIT1 FIN_WAIT2 "
               "TIME_WAIT CLOSE CLOSE_WAIT LAST_ACK LISTEN CLOSING"
               .split())
    def __init__(self):
        self.connections = None

    def _addr(self, part):
        host, port = part.split(':')
        port = str(int(port, 16))
        if len(host) == 8:
            parts = [host[0:2], host[2:4], host[4:6], host[6:8]]
            parts = [str(int(x, 16)) for x in parts]
            host = '.'.join(reversed(parts))
        else:
            host = "IPV6" # FIXME
        if host == '0.0.0.0':
            host = '*'
        elif host == '127.0.0.1':
            host = 'localhost'
        if port == '0':
            port = '*'
        return (host, port)

    def read_connections(self):
        def fromhex(x):
            return int(x, 16)
        self.connections = []
        for name in '/proc/net/tcp', '/proc/net/tcp6':
            f = open(name,'rt')
            headings = f.readline()
            for line in f:
                parts = line.split()
                if len(parts) < 10:
                    continue # broken line
                local = self._addr(parts[1])
                remote = self._addr(parts[2])
                state = self.TCP_STATE_NAMES[
                    fromhex(parts[3])-1]
                queueparts = parts[4].split(':')
                queues = tuple(map(fromhex,queueparts))
                uid, _timeout, inode = map(int, parts[7:10])
                if len(parts[1].split(":")[0]) == 8:
                    proto = "tcp"
                else:
                    proto = "tcp6"
                self.connections += [(proto, local, remote, state, queues, uid, inode)]

    def format_connections(self):
        t = "%-5s %6s %6s %15s:%-5s %15s:%-5s  %-11s  %s"
        headings = "Proto Send-Q Recv-Q Local Port Remote Port State User".split()
        return (t % tuple(headings) + '\n' +
            '\n'.join([t % (proto, rxq, txq, localh, localp, remoteh, remotep, state, getpwuid(uid)[0])
                   for proto, (localh, localp), (remoteh, remotep), state, (txq, rxq), uid, inode in self.connections
                   if localp == '6600' or remotep == '6600'
                   or getpwuid(uid)[0] == 'mpd']))

def update(label):
    # schedule next update
    GLib.timeout_add(1000, update, label)

    # don't update if not visible
    if not hasattr(label, "window") or not label.get_window().is_viewable():
        return

    netstat = Netstat()
    netstat.read_connections()
    netstats = netstat.format_connections()

    # XXX replace the shell commands with python code
    commands = [
        (_("Processes"), "ps wwu -C mpd".split()),
        (_("Files"), ["sh", "-c", "ls -ldh /etc/mpd.conf /var/lib/mpd "
                      "/var/lib/mpd/* /var/lib/mpd/*/*"]),
    ]
    outputs = [(title, subprocess.Popen(command,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE
                        ).communicate())
           for title, command in commands]

    sections = [outputs[0], (_("Networking"), (netstats, "")), outputs[1]]
    text = '\n'.join(["<b>%s</b>\n<tt>%s</tt><i>%s</i>\n" %
              (title, escape_html(stdout), escape_html(stderr))
              for title, (stdout, stderr) in sections])
    label.set_markup(text.decode(locale.getpreferredencoding(),
                     'replace'))

# nothing magical here, this constructs the parts of the tab when called:
def tab_construct():
    builder = ui.builder('localmpd', 'plugins')
    editbutton = builder.get_object('localmpd_edit_button')
    editbutton.connect('clicked', lambda *args: subprocess.Popen(
        ["gksu", "xdg-open", "/etc/mpd.conf"]))
    restartbutton = builder.get_object('localmpd_restart_button')
    restartbutton.connect('clicked', lambda *args:subprocess.Popen(
            ["gksu", "service", "mpd", "restart"]))

    label = builder.get_object('localmpd_data_label')
    update(label)

    window = builder.get_object('localmpd_scrolledwindow')
    window.show_all()
    tab_widget = builder.get_object('localmpd_tab_eventbox')

    # (tab content, icon name, tab name, the widget to focus on tab switch)
    return (window, tab_widget, "Local MPD", None)




# Copyright 2006-2009 Scott Horowitz <stonecrest@gmail.com>
# Copyright 2009-2014 Jonathan Ballet <jon@multani.info>
#
# This file is part of Sonata.
#
# Sonata is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Sonata is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Sonata.  If not, see <http://www.gnu.org/licenses/>.

"""
This module contains various constant definitions that any other
module can import without risk of cyclic imports.

Most of the constants are enum-like, that is, they provide symbolic
names for a set of values.

XXX Should some of these be moved to be private in some module, or
into config?

Example usage:
from consts import consts
...
if view == consts.VIEW_ALBUM: ...
"""

ART_LOCAL = 0
ART_LOCAL_REMOTE = 1
VIEW_FILESYSTEM = 0
VIEW_ARTIST = 1
VIEW_GENRE = 2
VIEW_ALBUM = 3
LYRIC_TIMEOUT = 10
NOTIFICATION_WIDTH_MAX = 500
NOTIFICATION_WIDTH_MIN = 350
FULLSCREEN_COVER_SIZE = 500
ART_LOCATION_HOMECOVERS = 0      # ~/.covers/[artist]-[album].jpg
ART_LOCATION_COVER = 1           # file_dir/cover.jpg
ART_LOCATION_ALBUM = 2           # file_dir/album.jpg
ART_LOCATION_FOLDER = 3          # file_dir/folder.jpg
ART_LOCATION_CUSTOM = 4          # file_dir/[custom]
ART_LOCATION_SINGLE = 6
ART_LOCATION_MISC = 7
ART_LOCATIONS_MISC = ['front.jpg', '.folder.jpg',
                      '.folder.png', 'AlbumArt.jpg',
                      'AlbumArtSmall.jpg']
LYRICS_LOCATION_HOME = 0         # ~/.lyrics/[artist]-[song].txt
LYRICS_LOCATION_PATH = 1         # file_dir/[artist]-[song].txt
LYRICS_LOCATION_HOME_ALT = 2     # ~/.lyrics/[artist] - [song].txt
LYRICS_LOCATION_PATH_ALT = 3     # file_dir/[artist] - [song].txt
LIB_COVER_SIZE = 32
COVERS_TYPE_STANDARD = 0
COVERS_TYPE_STYLIZED = 1
LIB_LEVEL_GENRE = 0
LIB_LEVEL_ARTIST = 1
LIB_LEVEL_ALBUM = 2
LIB_LEVEL_SONG = 3
NUM_ARTISTS_FOR_VA = 2
PREV_TRACK_RESTART = 5

# the names of the plug-ins that will be enabled by default
DEFAULT_PLUGINS = [
    'playlists',
    'streams',
    'lyricwiki',
    'localmpd',
]



# Copyright 2006-2009 Scott Horowitz <stonecrest@gmail.com>
# Copyright 2009-2014 Jonathan Ballet <jon@multani.info>
#
# This file is part of Sonata.
#
# Sonata is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Sonata is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Sonata.  If not, see <http://www.gnu.org/licenses/>.

import os

from gi.repository import Gtk, GLib

from sonata import ui, img


class TrayIconTips(Gtk.Window):
    """Custom tooltips derived from Gtk.Window() that allow for markup text and
    multiple widgets, e.g. a progress bar. ;)"""
    MARGIN = 4

    def __init__(self):
        super().__init__(type=Gtk.WindowType.POPUP)
        # from gtktooltips.c:gtk_tooltips_force_window
        self.set_app_paintable(True)
        self.set_resizable(False)
        self.set_name("gtk-tooltips")

        self.timer_tag = None
        self.notif_handler = None
        self.use_notifications_location = False
        self.notifications_location = 0

    def _calculate_pos(self, tray_icon):
        if tray_icon is not None:
            x, y, _, height = tray_icon.compute_pos()
        size = self.size_request()
        w = size.width
        h = size.height

        screen = self.get_screen()
        pointer_screen, px, py, _ = screen.get_display().get_pointer()
        if pointer_screen != screen:
            px = x
            py = y
        try:
            # Use the monitor that the systemtray icon is on
            monitor_num = screen.get_monitor_at_point(x, y)
        except:
            # No systemtray icon, use the monitor that the pointer is on
            monitor_num = screen.get_monitor_at_point(px, py)
        monitor = screen.get_monitor_geometry(monitor_num)

        try:
            # If the tooltip goes off the screen horizontally, realign it so
            # that it all displays.
            if (x + w) > monitor.x + monitor.width:
                x = monitor.x + monitor.width - w
            # If the tooltip goes off the screen vertically (i.e. the system
            # tray icon is on the bottom of the screen), realign the icon so
            # that it shows above the icon.
            if ((y + h + height + self.MARGIN) >
                monitor.y + monitor.height):
                y = y - h - self.MARGIN
            else:
                y = y + height + self.MARGIN
        except:
            pass

        if not self.use_notifications_location:
            try:
                return x, y
            except:
                #Fallback to top-left:
                return monitor.x, monitor.y
        elif self.notifications_location == 0:
            try:
                return x, y
            except:
                #Fallback to top-left:
                return monitor.x, monitor.y
        elif self.notifications_location == 1:
            return monitor.x, monitor.y
        elif self.notifications_location == 2:
            return monitor.x + monitor.width - w, monitor.y
        elif self.notifications_location == 3:
            return monitor.x, monitor.y + monitor.height - h
        elif self.notifications_location == 4:
            return monitor.x + monitor.width - w, \
                    monitor.y + monitor.height - h
        elif self.notifications_location == 5:
            return monitor.x + (monitor.width - w) / 2, \
                    monitor.y + (monitor.height - h) / 2

    def _start_delay(self, tray_icon):
        self.timer_tag = GLib.timeout_add(500, self._tips_timeout, tray_icon)

    def _tips_timeout(self, tray_icon):
        self.use_notifications_location = False
        self._real_display(tray_icon)

    def _remove_timer(self):
        self.hide()
        if self.timer_tag:
            GLib.source_remove(self.timer_tag)
        self.timer_tag = None

    def _real_display(self, tray_icon):
        x, y = self._calculate_pos(tray_icon)
        self.move(x, y)
        self.show()

    # Public API

    def hide(self):
        Gtk.Window.hide(self)
        self.notif_handler = None

    def add_widget(self, widget_to_add):
        self.add(widget_to_add)


class TrayIcon:
    """Tray icon which use Gtk.StatusIcon"""

    def __init__(self, window, traymenu, traytips):
        self.statusicon = None
        self.window = window
        self.traymenu = traymenu
        self.traytips = traytips

    def compute_pos(self):
        _ok, _screen, rect, _orient = self.statusicon.get_geometry()
        return (rect.x, rect.y, rect.height, rect.width)

    def initialize(self, on_click, on_scroll):
        self.statusicon = Gtk.StatusIcon()
        self.statusicon.connect('button-press-event', on_click)
        self.statusicon.connect('scroll-event', on_scroll)
        GLib.timeout_add(250, self._iterate_status_icon)

    def is_visible(self):
        """Visible and/or notification activated"""
        return self.statusicon.is_embedded() and \
                self.statusicon.get_visible()

    def update_icon(self, icon_name):
        # FIXME (why doesn't this work instead?):
        # self.statusicon.set_from_stock(icon_name)
        context = self.traymenu.get_style_context()
        icon = Gtk.IconFactory.lookup_default(icon_name)
        icon_pb = icon.render_icon_pixbuf(context, -1)
        self.statusicon.set_from_pixbuf(icon_pb)

    def show(self):
        self.statusicon.set_visible(True)

    def hide(self):
        self.statusicon.set_visible(False)

    def is_available(self):
        return self.statusicon.is_embedded()

    def _iterate_status_icon(self):
        if self.is_visible():
            self._tooltip_show_manually()
        GLib.timeout_add(250, self._iterate_status_icon)

    def _tooltip_show_manually(self):
        # Since there is no signal to connect to when the user puts their
        # mouse over the trayicon, we will check the mouse position
        # manually and show/hide the window as appropriate. This is called
        # every iteration. Note: This should not occur if self.traytips.notif_
        # handler has a value, because that means that the tooltip is already
        # visible, and we don't want to override that setting simply because
        # the user's cursor is not over the tooltip.
        if self.traymenu.get_property('visible') and \
           self.traytips.notif_handler != -1:
            self.traytips._remove_timer()
        elif not self.traytips.notif_handler:
            _pscreen, px, py, _mods = \
                    self.window.get_screen().get_display().get_pointer()
            x, y, width, height = self.compute_pos()
            if px >= x and px <= x + width and py >= y and py <= y + height:
                self.traytips._start_delay(self)
            else:
                self.traytips._remove_timer()



#!/usr/bin/env python
# Copyright 2006-2009 Scott Horowitz <stonecrest@gmail.com>
# Copyright 2009-2014 Jonathan Ballet <jon@multani.info>
#
# This file is part of Sonata.
#
# Sonata is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Sonata is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Sonata.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gdk, Pango


class CrumbButton(Gtk.ToggleButton):
    """A ToggleButton tailored for use as a breadcrumb."""

    def __init__(self, image, label):
        super().__init__()

        self.set_properties(can_focus=False, relief=Gtk.ReliefStyle.NONE)
        self.label = label

        # adapt Gtk.Button internal layout code:
        hbox = Gtk.HBox(spacing=2)
        align = Gtk.Alignment.new(0.5, 0.5, 0, 0)
        hbox.pack_start(image, False, False, 0)
        hbox.pack_start(label, True, True, 0)
        align.add(hbox)

        self.add(align)

    def ellipsize(self, do_ellipsize=True):
        can_ellipsize = len(self.label.get_text()) >= 15
        if do_ellipsize and can_ellipsize:
            mode = Pango.EllipsizeMode.END
            width_chars = 15
        else:
            mode = Pango.EllipsizeMode.NONE
            width_chars = -1

        self.label.set_ellipsize(mode)
        self.label.set_max_width_chars(width_chars)
        self.label.set_width_chars(width_chars)
        return do_ellipsize and can_ellipsize


class CrumbBox(Gtk.Box):
    """A box layout similar to Gtk.HBox, but specifically for breadcrumbs.

    * Crumbs in the middle are replaced with an ellipsis if necessary.
    * The root, parent, and current element are always kept visible.
    * The root and parent are put in a condensed form if necessary.
    * The current element is truncated if necessary.
    """
    __gtype_name__ = 'CrumbBox'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def set_crumb_break(self, widget):
        self.crumb_break = widget

    def do_get_preferred_width(self):
        """This gets called to determine the size we request"""
        reqs = [w.size_request() for w in self]

        # Request "minimum" size:
        heights = [r.height for r in reqs]
        height = max([0] + heights)
        widths = [r.width for r in reqs]
        width = max([0] + widths)
        natural_width = self._req_sum(reqs)

        if len(reqs) == 0: # empty
            width = 0
        elif len(reqs) < 3: # current crumb
            width = natural_width
        elif len(reqs) == 3: # parent and current
            width = 2 * height + self.props.spacing
        elif len(reqs) == 4: # root, parent and current
            width = 3 * height + 2 * self.props.spacing
        elif len(reqs) > 4: # root, break, parent, current
            pad = 3 * self.props.spacing
            width = 3 * height + reqs[1].width + pad

        return width, natural_width

    def _req_sum(self, reqs):
        pad = 0 if not reqs else (len(reqs)-1) * self.props.spacing
        return pad + sum([req.width for req in reqs])

    def _condense(self, req, w):
# FIXME show and hide cause a fight in an infinite loop
#		try:
#			w.get_child().get_child().get_children()[1].hide()
#		except (AttributeError, IndexError):
#			pass
        return req # XXX simplistic: set square size for now

    def _uncondense(self, w):
        pass
#		try:
#			w.get_child().get_child().get_children()[1].show()
#		except (AttributeError, IndexError):

    def _truncate(self, req, amount):
        req.width -= amount
        if req.width < 0:
            req.width = 0
        return req

    def do_size_allocate(self, allocation):
        """This gets called to layout our child widgets"""
        # XXX allocation seems to stick at some width, no matter how much wider
        # the window gets, but if we go down the path and back, it goes higher?
        # Gtk bug?
        x0, y0 = allocation.x, allocation.y
        w0, h0 = allocation.width, allocation.height

        crumbs = self.get_children()

        if len(crumbs) < 2:
            if self.crumb_break:
                self.crumb_break.hide()
            Gtk.Box.do_size_allocate(self, allocation)
            return

        # Undo any earlier condensing and ellipsizing
        for crumb in crumbs:
            crumb.ellipsize(False)
        if len(crumbs) > 2:
            self._uncondense(crumbs[-2])

        reqs = [w.get_child_requisition() for w in crumbs]

        # Step 1: Try ellipsizing
        if self._req_sum(reqs) > w0:
            # We want a descending sort by widths
            # We ellipsize biggest to smallest to preserve the most crumbs
            crumbsorted = [(w.width, crumb) for w, crumb in zip(reqs, crumbs)]
            sorted(crumbsorted, key=lambda crumb: crumb[0], reverse=True)
            for x, crumb in crumbsorted:
                if crumb.ellipsize():
                    i = crumbs.index(crumb)
                    reqs[i] = crumb.get_child_requisition()
                    if self._req_sum(reqs) < w0:
                        break

        # Step 2: remove as many crumbs as needed except:
        # Never remove the parent or current
        hidden = []
        while self._req_sum(reqs) > w0:
            if len(crumbs) > 2:
                    hidden.append(crumbs.pop(0))
                    reqs.pop(0)
            else:
                break # don't remove the parent

        # If necessary, condense the parent crumb
        if self._req_sum(reqs) > w0 and len(crumbs) > 2:
            reqs[-2] = self._condense(reqs[-2], crumbs[-2])

        # If necessary, truncate the current crumb
        if self._req_sum(reqs) > w0:
            reqs[-1] = self._truncate(reqs[-1], self._req_sum(reqs) - w0)
            # Now we are at minimum width

        # Only show the break (ellipsis) if we have hidden crumbs
        if self.crumb_break:
            if len(hidden):
                self.crumb_break.show()
            else:
                self.crumb_break.hide()

        x = 0
        for w, req in zip(crumbs, reqs):
            alloc = Gdk.Rectangle()
            alloc.x = x0 + x
            alloc.y = y0
            alloc.width = req.width
            alloc.height = h0
            w.size_allocate(alloc)
            w.show()
            x += req.width + self.props.spacing

        for w in hidden:
            alloc = Gdk.Rectangle()
            alloc.x = -1
            alloc.y = -1
            alloc.width = 0
            alloc.height = 0
            w.size_allocate(alloc)
            w.hide()

        Gtk.Box.do_size_allocate(self, allocation)

# this file can be run as a simple test program:
# FIXME or XXX this is out of sync with the rest of file
if __name__ == '__main__':
    w = Gtk.Window()
    crumbs = CrumbBox(spacing=2)

    items = [
        (Gtk.STOCK_HARDDISK, "Filesystem"),
        (None, None), # XXX for ellipsis
        (Gtk.STOCK_OPEN, "home"),
        (Gtk.STOCK_OPEN, "user"),
        (Gtk.STOCK_OPEN, "music"),
        (Gtk.STOCK_OPEN, "genre"),
        (Gtk.STOCK_OPEN, "artist"),
        (Gtk.STOCK_OPEN, "album"),
        ]
    for stock, text in items:
        if stock:
            image = Gtk.Image.new_from_stock(stock,
                             Gtk.IconSize.MENU)
            crumbs.pack_start(CrumbButton(image, Gtk.Label(label=text)), False, False, 0)
        else:
            crumbs.pack_start(Gtk.Label("..."), False, False, 0)

    w.add(crumbs)
    w.connect('hide', Gtk.main_quit)
    w.show_all()

    Gtk.main()



#!/usr/bin/env python

### BEGIN PLUGIN INFO
# [plugin]
# plugin_format: 0, 0
# name: Gajim tune
# version: 0, 0, 1
# description: Update the MPRIS (in Gajim etc.) tune information.
# author: Fomin Denis
# author_email: fominde@gmail.com
# url: http://www.nongnu.org/sonata/
# license: GPL v3 or later
# [capabilities]
# enablables: on_enable
# playing_song_observers: on_song_change
### END PLUGIN INFO

import dbus.service

songlabel = None
lasttune = ''
tune = None

# this gets called when the plugin is loaded, enabled, or disabled:
def on_enable(state):
    global tune
    if state and not tune:
        tune = mpdtune()
    else:
        if tune:
            title = artist = album = ''
            tune.TrackChange(dbus.Dictionary(
                        {'title' : title, 'artist' : artist, 'album' : album}
                        ))


def on_song_change(songinfo):
    global tune, lasttune
    if lasttune == songinfo:
        return
    lasttune = songinfo
    title = artist = album = ''

    if not songinfo:
        # mpd stopped
        if tune:
            tune.TrackChange(dbus.Dictionary(
                        {'title' : title, 'artist' : artist, 'album' : album}))
        return

    if 'title' in songinfo:
        title = songinfo['title']
    if 'artist' in songinfo:
        artist = songinfo['artist']
    if 'album' in songinfo:
        album = songinfo['album']
    if 'name' in songinfo and artist =='':
        artist = songinfo['name']
    if 'file' in songinfo and title =='':
        album = songinfo['file']
        if not album.startswith('http:'):
            title = album.rpartition('/')[2]
    if tune:
        tune.TrackChange(dbus.Dictionary(
                    {'title' : title, 'artist' : artist, 'album' : album}
                    ))

class mpdtune(dbus.service.Object):
    def __init__(self):
        super().__init__(dbus.SessionBus(), '/Player')

    @dbus.service.signal(dbus_interface = 'org.freedesktop.MediaPlayer')
    def TrackChange(self, trackinfo):
        return



# Copyright 2006-2009 Scott Horowitz <stonecrest@gmail.com>
# Copyright 2009-2014 Jonathan Ballet <jon@multani.info>
#
# This file is part of Sonata.
#
# Sonata is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Sonata is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Sonata.  If not, see <http://www.gnu.org/licenses/>.

"""This module implements the format strings used to display song info.

Example usage:
import formatting
colnames = formatting.parse_colnames(self.config.currentformat)
...
newtitle = formatting.parse(self.config.titleformat, self.songinfo,
                            False, True)
...
formatcodes = formatting.formatcodes
"""

import re
import os

from sonata import mpdhelper as mpdh
from sonata import misc


class FormatCode:
    """Implements deafult format code behavior.

    Replaces all instances of %code with the value of key or default if the
    key doesn't exist.
    """

    def __init__(self, code, description, column, key, default=_("Unknown")):
        self.code = code
        self.description = description
        self.column = column
        self.key = key
        self.default = default

    def format(self, item, wintitle, songpos):
        """Returns the value used in place of the format code"""
        return str(item.get(self.key, self.default))


class NumFormatCode(FormatCode):
    """Implements format code behavior for numeric values.

    Used for numbers which need special padding.
    """

    def __init__(self, code, description, column, key, default, padding):
        super().__init__(code, description, column, key, default)
        self.padding = padding

    def format(self, item, wintitle, songpos):
        return str(item.get(self.key, self.default)).zfill(self.padding)


class PathFormatCode(FormatCode):
    """Implements format code behavior for path values."""

    def __init__(self, code, description, column, key, path_func):
        """

        path_func: os.path function to apply
        """
        super().__init__(code, description, column, key)
        self.func = getattr(os.path, path_func)

    def format(self, item, wintitle, songpos):
        return self.func(FormatCode.format(self, item, wintitle,
                            songpos))


class TitleFormatCode(FormatCode):
    """Implements format code behavior for track titles."""

    def format(self, item, wintitle, songpos):
        path = item['file']
        full_path = re.match(r"^(http://|ftp://)", path)
        self.default = path if full_path else os.path.basename(path)
        self.default = misc.escape_html(self.default)
        return FormatCode.format(self, item, wintitle, songpos)


class LenFormatCode(FormatCode):
    """Implements format code behavior for song length."""

    def format(self, item, wintitle, songpos):
        time = FormatCode.format(self, item, wintitle, songpos)
        if time.isdigit():
            time = misc.convert_time(int(time))
        return time


class ElapsedFormatCode(FormatCode):
    """Implements format code behavior for elapsed time."""

    def format(self, item, wintitle, songpos):
        if not wintitle:
            return "%E"
        elapsed_time = songpos.split(':')[0] if songpos else self.default
        if elapsed_time.isdigit():
            elapsed_time = misc.convert_time(int(elapsed_time))
        return elapsed_time

formatcodes = [FormatCode('A', _('Artist name'), _("Artist"), 'artist'),
           FormatCode('B', _('Album name'), _("Album"), 'album'),
           TitleFormatCode('T', _('Track name'), _("Track"), 'title'),
           NumFormatCode('N', _('Track number'), _("#"), 'track', '00', 2),
           NumFormatCode('D', _('Disc number'), _("#"), 'disc', '0', 0),
           FormatCode('Y', _('Year'), _("Year"), 'date', '?'),
           FormatCode('G', _('Genre'), _("Genre"), 'genre'),
           PathFormatCode('P', _('File path'), _("Path"), 'file',
                'dirname'),
           PathFormatCode('F', _('File name'), _("File"), 'file',
                'basename'),
           FormatCode('S', _('Stream name'), _("Stream"), 'name'),
           LenFormatCode('L', _('Song length'), _("Len"), 'time', '?'),
           ElapsedFormatCode('E', _('Elapsed time (title only)'), None,
                 'songpos', '?')]

replace_map = dict((code.code, code) for code in formatcodes)
replace_expr = r"%%[%s]" % "".join(k for k in replace_map.keys())


def _return_substrings(format):
    """Split format along the { and } characters.

    For example: %A{-%T} {%L} -> ['%A', '{-%T} ', '{%L}']"""
    substrings = []
    end = format
    while len(end) > 0:
        begin, sep1, end = end.partition('{')
        substrings.append(begin)
        if len(end) == 0:
            substrings.append(sep1)
            break
        begin, sep2, end = end.partition('}')
        substrings.append(sep1 + begin + sep2)
    return substrings


def parse_colnames(format):

    def replace_format(m):
        format_code = replace_map.get(m.group(0)[1:])
        return format_code.column

    cols = [re.sub(replace_expr, replace_format, s).
        replace("{", "").
        replace("}", "").
        # If the user wants the format of, e.g., "#%N", we'll
        # ensure the # doesn't show up twice in a row.
        replace("##", "#")
        for s in format.split('|')]
    return cols


class EmptyBrackets(Exception):
    pass


def _format_substrings(text, item, wintitle, songpos):
    has_brackets = text.startswith("{") and text.endswith("}")

    def formatter(m):
        format_code = replace_map[m.group(0)[1:]]
        if has_brackets and format_code.key not in item:
            raise EmptyBrackets
        return format_code.format(item, wintitle, songpos)

    try:
        text = re.sub(replace_expr, formatter, text)
    except EmptyBrackets:
        return ""

    return text[1:-1] if has_brackets else text


def parse(format, item, use_escape_html, wintitle=False, songpos=None):
    substrings = _return_substrings(format)
    text = "".join(_format_substrings(sub, item, wintitle, songpos)
            for sub in substrings)
    return misc.escape_html(text) if use_escape_html else text



# Copyright 2006-2009 Scott Horowitz <stonecrest@gmail.com>
# Copyright 2009-2014 Jonathan Ballet <jon@multani.info>
#
# This file is part of Sonata.
#
# Sonata is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Sonata is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Sonata.  If not, see <http://www.gnu.org/licenses/>.

from collections import namedtuple

_SongRecord = namedtuple('SongRecord',
                        ['album', 'artist', 'genre', 'year', 'path'])

def SongRecord(album=None, artist=None, genre=None, year=None, path=None):
    return _SongRecord(album, artist, genre, year, path)



__license__ = """
Sonata, an elegant GTK+ client for the Music Player Daemon

Copyright 2006-2009 Scott Horowitz <stonecrest@gmail.com>
Copyright 2009-2014 Jonathan Ballet <jon@multani.info>

This file is part of Sonata.

Sonata is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Sonata is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Sonata.  If not, see <http://www.gnu.org/licenses/>.
"""



# Copyright 2006-2009 Scott Horowitz <stonecrest@gmail.com>
# Copyright 2009-2014 Jonathan Ballet <jon@multani.info>
#
# This file is part of Sonata.
#
# Sonata is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Sonata is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Sonata.  If not, see <http://www.gnu.org/licenses/>.

"""
This module provides a user interface for changing configuration
variables.

Example usage:
import preferences
...
prefs = preferences.Preferences()
prefs.on_prefs_real(self.window, self.prefs_window_response, tab callbacks...)
"""

import gettext, hashlib

from gi.repository import Gtk, Gdk, GdkPixbuf

from sonata.config import Config
from sonata.pluginsystem import pluginsystem
from sonata import ui, misc, consts, formatting
import os

class Extras_cbs:
    """Callbacks and data specific to the extras tab"""
    popuptimes = []
    notif_toggled = None
    crossfade_changed = None
    crossfade_toggled = None

class Display_cbs:
    """Callbacks specific to the display tab"""
    stylized_toggled = None
    art_toggled = None
    playback_toggled = None
    progress_toggled = None
    statusbar_toggled = None
    lyrics_toggled = None
    trayicon_available = None

class Behavior_cbs:
    """Callbacks and data specific to the behavior tab"""
    trayicon_toggled = None
    trayicon_in_use = None
    sticky_toggled = None
    ontop_toggled = None
    decorated_toggled = None
    infofile_changed = None

class Format_cbs:
    """Callbacks specific to the format tab"""
    currentoptions_changed = None
    libraryoptions_changed = None
    titleoptions_changed = None
    currsongoptions1_changed = None
    currsongoptions2_changed = None

class Preferences():
    """This class implements a preferences dialog for changing
    configuration variables.

    Many changes are applied instantly with respective
    callbacks. Closing the dialog causes a response callback.
    """
    def __init__(self, config, reconnect, renotify, reinfofile,
             settings_save, populate_profiles_for_menu):

        self.config = config

        # These are callbacks to Main
        self.reconnect = reconnect
        self.renotify = renotify
        self.reinfofile = reinfofile
        self.settings_save = settings_save
        self.populate_profiles_for_menu = populate_profiles_for_menu

        # Temporary flag:
        self.updating_nameentry = False

        self.prev_host = None
        self.prev_password = None
        self.prev_port = None

        self.window = None
        self.last_tab = 0
        self.display_trayicon = None
        self.direntry = None
        self.using_mpd_env_vars = False

    def on_prefs_real(self):
        """Display the preferences dialog"""

        self.builder = ui.builder('preferences')
        self.provider = ui.css_provider('preferences')
        self.prefswindow = self.builder.get_object('preferences_dialog')
        self.prefswindow.set_transient_for(self.window)
        self.prefsnotebook = self.builder.get_object('preferences_notebook')

        tabs = ('mpd', 'display', 'behavior', 'format', 'extras', 'plugins')

        for name in tabs:
            func = getattr(self, '%s_tab' % name)
            cbs = globals().get('%s_cbs' % name.capitalize())
            func(cbs)

        close_button = self.builder.get_object('preferences_closebutton')
        self.prefswindow.show_all()
        self.prefsnotebook.set_current_page(self.last_tab)
        close_button.grab_focus()
        self.prefswindow.connect('response', self._window_response)
        # Save previous connection properties to determine if we should try to
        # connect to MPD after prefs are closed:
        self.prev_host = self.config.host[self.config.profile_num]
        self.prev_port = self.config.port[self.config.profile_num]
        self.prev_password = self.config.password[self.config.profile_num]

    def mpd_tab(self, cbs=None):
        """Construct and layout the MPD tab"""
        #frame.set_shadow_type(Gtk.ShadowType.NONE)
        controlbox = self.builder.get_object('connection_frame_label_widget')
        profiles = self.builder.get_object('connection_profiles')
        add_profile = self.builder.get_object('connection_add_profile')
        remove_profile = self.builder.get_object('connection_remove_profile')
        self._populate_profile_combo(profiles, self.config.profile_num,
            remove_profile)
        nameentry = self.builder.get_object('connection_name')
        hostentry = self.builder.get_object('connection_host')
        portentry = self.builder.get_object('connection_port')
        direntry = self.builder.get_object('connection_dir')
        self.direntry = direntry
        direntry.connect('selection-changed', self._direntry_changed,
            profiles)
        passwordentry = self.builder.get_object('connection_password')
        autoconnect = self.builder.get_object('connection_autoconnect')
        autoconnect.set_active(self.config.autoconnect)
        autoconnect.connect('toggled', self._config_widget_active,
            'autoconnect')
        # Fill in entries with current profile:
        self._profile_chosen(profiles, nameentry, hostentry,
            portentry, passwordentry, direntry)
        # Update display if $MPD_HOST or $MPD_PORT is set:
        host, port, password = misc.mpd_env_vars()
        if host or port:
            self.using_mpd_env_vars = True
            if not host:
                host = ""
            if not port:
                port = 0
            if not password:
                password = ""
            hostentry.set_text(str(host))
            portentry.set_value(port)
            passwordentry.set_text(str(password))
            nameentry.set_text(_("Using MPD_HOST/PORT"))
            for widget in [hostentry, portentry, passwordentry,
                       nameentry, profiles, add_profile,
                       remove_profile]:
                widget.set_sensitive(False)
        else:
            self.using_mpd_env_vars = False
            nameentry.connect('changed', self._nameentry_changed,
                profiles, remove_profile)
            hostentry.connect('changed', self._hostentry_changed,
                profiles)
            portentry.connect('value-changed',
                self._portentry_changed, profiles)
            passwordentry.connect('changed',
                self._passwordentry_changed, profiles)
            profiles.connect('changed',
                self._profile_chosen, nameentry, hostentry,
                portentry, passwordentry, direntry)
            add_profile.connect('clicked', self._add_profile,
                nameentry, profiles, remove_profile)
            remove_profile.connect('clicked', self._remove_profile,
                profiles, remove_profile)

    def extras_tab(self, cbs):
        """Construct and layout the extras tab"""
        if not self.scrobbler.imported():
            self.config.as_enabled = False

        as_checkbox = self.builder.get_object('scrobbler_check')
        as_checkbox.set_active(self.config.as_enabled)
        as_user_label = self.builder.get_object('scrobbler_username_label')
        as_user_entry = self.builder.get_object('scrobbler_username_entry')
        as_user_entry.set_text(self.config.as_username)
        as_user_entry.connect('changed', self._as_username_changed)
        as_pass_label = self.builder.get_object('scrobbler_password_label')
        as_pass_entry = self.builder.get_object('scrobbler_password_entry')
        as_pass_entry.set_text(self.config.as_password_md5)
        as_pass_entry.connect('changed', self._as_password_changed)
        display_notification = self.builder.get_object('notification_check')
        display_notification.set_active(self.config.show_notification)

        time_names = ["%s %s" %
            (i , ngettext('second', 'seconds', int(i)))
            for i in cbs.popuptimes if i != _('Entire song')]
        notification_options = self.builder.get_object('notification_time_combo')
        for time in time_names:
            notification_options.append_text(time)
        notification_options.connect('changed', self._notiftime_changed)
        notification_options.set_active(self.config.popup_option)
        notification_locs = self.builder.get_object('notification_loc_combo')
        notification_locs.set_active(self.config.traytips_notifications_location)
        notification_locs.connect('changed', self._notiflocation_changed)
        notifhbox = self.builder.get_object('notification_box')
        display_notification.connect('toggled', cbs.notif_toggled,
            notifhbox)
        if not self.config.show_notification:
            notifhbox.set_sensitive(False)

        crossfadespin = self.builder.get_object('crossfade_time')
        crossfadespin.set_value(self.config.xfade)
        crossfadespin.connect('value-changed', cbs.crossfade_changed)
        crossfadelabel2 = self.builder.get_object('crossfade_label')
        crossfadelabel3 = self.builder.get_object('crossfade_extra_label')
        crossfadecheck = self.builder.get_object('crossfade_check')
        crossfadecheck.connect('toggled',
            self._crossfadecheck_toggled, crossfadespin,
            crossfadelabel2, crossfadelabel3)
        crossfadecheck.connect('toggled', cbs.crossfade_toggled,
            crossfadespin)
        crossfadecheck.set_active(self.config.xfade_enabled)
        crossfadecheck.toggled() # Force the toggled callback

        as_checkbox.connect('toggled', self._as_enabled_toggled,
            as_user_entry, as_pass_entry, as_user_label,
            as_pass_label)
        if not self.config.as_enabled or not self.scrobbler.imported():
            for widget in (as_user_entry, as_pass_entry,
                    as_user_label, as_pass_label):
                widget.set_sensitive(False)

    def display_tab(self, cbs):
        """Construct and layout the display tab"""

        art = self.builder.get_object('art_check')
        art.set_active(self.config.show_covers)
        stylized_combo = self.builder.get_object('art_style_combo')
        stylized_combo.set_active(self.config.covers_type)
        stylized_combo.connect('changed', cbs.stylized_toggled)
        art_prefs = self.builder.get_object('art_preferences')
        art_prefs.set_sensitive(self.config.show_covers)
        art_combo = self.builder.get_object('art_search_combo')
        art_combo.set_active(self.config.covers_pref)
        art_combo.connect('changed', self._config_widget_active, 'covers_pref')

        #FIXME move into preferences_display.ui?
        art_location = self.builder.get_object('art_save_combo')
        for item in ["%s/%s" % (_("SONG_DIR"), item)
            for item in ("cover.jpg", "album.jpg", "folder.jpg",
                self.config.art_location_custom_filename or _("custom"))]:
            art_location.append_text(item)
        art_location.set_active(self.config.art_location)
        art_location.connect('changed', self._art_location_changed)

        art.connect('toggled', cbs.art_toggled, art_prefs)
        playback = self.builder.get_object('playback_buttons_check')
        playback.set_active(self.config.show_playback)
        playback.connect('toggled', cbs.playback_toggled)
        progress = self.builder.get_object('progressbar_check')
        progress.set_active(self.config.show_progress)
        progress.connect('toggled', cbs.progress_toggled)
        statusbar = self.builder.get_object('statusbar_check')
        statusbar.set_active(self.config.show_statusbar)
        statusbar.connect('toggled', cbs.statusbar_toggled)
        lyrics = self.builder.get_object('lyrics_check')
        lyrics.set_active(self.config.show_lyrics)
        lyrics_location = self.builder.get_object('lyrics_save_combo')
        lyrics_location.set_active(self.config.lyrics_location)
        lyrics_location.connect('changed', self._lyrics_location_changed)
        lyrics_location_hbox = self.builder.get_object('lyrics_preferences')
        lyrics_location_hbox.set_sensitive(self.config.show_lyrics)
        lyrics.connect('toggled', cbs.lyrics_toggled, lyrics_location_hbox)
        trayicon = self.builder.get_object('tray_icon_check')
        self.display_trayicon = trayicon
        trayicon.set_active(self.config.show_trayicon)
        trayicon.set_sensitive(cbs.trayicon_available)

    def behavior_tab(self, cbs):
        """Construct and layout the behavior tab"""

        frame = self.builder.get_object('behavior_frame')
        sticky = self.builder.get_object('behavior_sticky_check')
        sticky.set_active(self.config.sticky)
        sticky.connect('toggled', cbs.sticky_toggled)
        ontop = self.builder.get_object('behavior_ontop_check')
        ontop.set_active(self.config.ontop)
        ontop.connect('toggled', cbs.ontop_toggled)
        decor = self.builder.get_object('behavior_decor_check')
        decor.set_active(not self.config.decorated)
        decor.connect('toggled', cbs.decorated_toggled, self.prefswindow)
        minimize = self.builder.get_object('behavior_minimize_check')
        minimize.set_active(self.config.minimize_to_systray)
        minimize.connect('toggled', self._config_widget_active,
            'minimize_to_systray')
        self.display_trayicon.connect('toggled', cbs.trayicon_toggled,
            minimize)
        minimize.set_sensitive(cbs.trayicon_in_use)

        update_start = self.builder.get_object('misc_updatestart_check')
        update_start.set_active(self.config.update_on_start)
        update_start.connect('toggled', self._config_widget_active,
            'update_on_start')
        exit_stop = self.builder.get_object('misc_exit_stop_check')
        exit_stop.set_active(self.config.stop_on_exit)
        exit_stop.connect('toggled', self._config_widget_active,
            'stop_on_exit')
        infofile_usage = self.builder.get_object('misc_infofile_usage_check')
        infofile_usage.set_active(self.config.use_infofile)
        infopath_options = self.builder.get_object('misc_infofile_entry')
        infopath_options.set_text(self.config.infofile_path)
        infopath_options.connect('focus_out_event',
                    cbs.infofile_changed)
        infopath_options.connect('activate', cbs.infofile_changed, None)
        if not self.config.use_infofile:
            infopath_options.set_sensitive(False)
        infofile_usage.connect('toggled', self._infofile_toggled,
            infopath_options)

    def format_tab(self, cbs):
        """Construct and layout the format tab"""

        playlist_entry = self.builder.get_object('format_playlist_entry')
        playlist_entry.set_text(self.config.currentformat)
        library_entry = self.builder.get_object('format_library_entry')
        library_entry.set_text(self.config.libraryformat)
        window_entry = self.builder.get_object('format_window_entry')
        window_entry.set_text(self.config.titleformat)
        current1_entry = self.builder.get_object('format_current1_entry')
        current1_entry.set_text(self.config.currsongformat1)
        current2_entry = self.builder.get_object('format_current2_entry')
        current2_entry.set_text(self.config.currsongformat2)
        entries = [playlist_entry, library_entry, window_entry, current1_entry,
                   current2_entry]

        entry_cbs = (cbs.currentoptions_changed,
                 cbs.libraryoptions_changed,
                 cbs.titleoptions_changed,
                 cbs.currsongoptions1_changed,
                 cbs.currsongoptions2_changed)
        for entry, cb, next in zip(entries, entry_cbs,
                entries[1:] + entries[:1]):
            entry.connect('focus_out_event', cb)
            entry.connect('activate', lambda _, n: n.grab_focus(),
                    next)

        format_grid = self.builder.get_object('format_avail_descs')
        codeset = formatting.formatcodes
        codes = (codeset[:len(codeset) // 2],
                 codeset[len(codeset) // 2:])

        def make_label(content):
            return Gtk.Label('<small>{}</small>'.format(content),
                             use_markup=True, xalign=0, yalign=0)

        for column_base, codegroup in enumerate(codes):
            column = column_base * 2
            for row, code in enumerate(codegroup):
                format_code = make_label("%{}".format(code.code))
                format_code.get_style_context().add_class('format_code')
                format_desc = make_label(code.description)
                format_grid.attach(format_code, column, row, 1, 1)
                format_grid.attach(format_desc, column + 1, row, 1, 1)

        additionalinfo = self.builder.get_object('format_additional_label')
        # FIXME need to either separate markup from localized strings OR
        # include markup in the strings and let the translators work around them
        row = len(codes[0])
        enclosed_code = make_label('{ }')
        enclosed_code.get_style_context().add_class('format_code')
        enclosed_desc = make_label(
            _('Info displayed only if all enclosed tags are defined'))
        enclosed_desc.set_line_wrap(True)
        enclosed_desc.set_max_width_chars(40)
        column_code = make_label('|')
        column_code.get_style_context().add_class('format_code')
        column_desc = make_label(_('Creates columns in the current playlist'))
        column_desc.set_line_wrap(True)
        column_desc.set_max_width_chars(40)

        for widget in (enclosed_code, enclosed_desc):
            widget.get_style_context().add_class('additional_format')
        format_grid.attach(enclosed_code, 0, row, 1, 1)
        format_grid.attach(enclosed_desc, 1, row, 3, 1)
        row += 1
        format_grid.attach(column_code, 0, row, 1, 1)
        format_grid.attach(column_desc, 1, row, 3, 1)

    def plugins_tab(self, cbs=None):
        """Construct and layout the plugins tab"""

        self.plugin_UIManager = self.builder.get_object('plugins_ui_manager')
        menu_handlers = {
            "plugin_configure": self.plugin_configure,
            "plugin_about": self.plugin_about
        }
        self.builder.connect_signals(menu_handlers)

        self.pluginview = self.builder.get_object('plugins_treeview')
        self.pluginselection = self.pluginview.get_selection()
        plugindata = self.builder.get_object('plugins_store')
        self.pluginview.connect('button-press-event', self.plugin_click)

        plugincheckcell = self.builder.get_object('plugins_check_renderer')
        plugincheckcell.connect('toggled', self.plugin_toggled,
            (plugindata, 0))

        plugindata.clear()
        for plugin in pluginsystem.get_info():
            pb = self.plugin_get_icon_pixbuf(plugin)
            plugin_text = "<b>" + plugin.longname + "</b> " + plugin.version_string
            plugin_text += "\n" + plugin.description
            enabled = plugin.get_enabled()
            plugindata.append((enabled, pb, plugin_text))

    def _window_response(self, window, response):
        if response == Gtk.ResponseType.CLOSE:
            self.last_tab = self.prefsnotebook.get_current_page()
            #XXX: These two are probably never triggered
            if self.config.show_lyrics and self.config.lyrics_location != consts.LYRICS_LOCATION_HOME:
                if not os.path.isdir(self.config.current_musicdir):
                    ui.show_msg(self.window, _("To save lyrics to the music file's directory, you must specify a valid music directory."), _("Music Dir Verification"), 'musicdirVerificationError', Gtk.ButtonsType.CLOSE)
                    # Set music_dir entry focused:
                    self.prefsnotebook.set_current_page(0)
                    self.direntry.grab_focus()
                    return
            if self.config.show_covers and self.config.art_location != consts.ART_LOCATION_HOMECOVERS:
                if not os.path.isdir(self.config.current_musicdir):
                    ui.show_msg(self.window, _("To save artwork to the music file's directory, you must specify a valid music directory."), _("Music Dir Verification"), 'musicdirVerificationError', Gtk.ButtonsType.CLOSE)
                    # Set music_dir entry focused:
                    self.prefsnotebook.set_current_page(0)
                    self.direntry.grab_focus()
                    return
            if not self.using_mpd_env_vars:
                if self.prev_host != self.config.host[self.config.profile_num] or self.prev_port != self.config.port[self.config.profile_num] or self.prev_password != self.config.password[self.config.profile_num]:
                    # Try to connect if mpd connection info has been updated:
                    ui.change_cursor(Gdk.Cursor.new(Gdk.CursorType.WATCH))
                    self.reconnect()
            self.settings_save()
            self.populate_profiles_for_menu()
            ui.change_cursor(None)
        window.destroy()

    def _config_widget_active(self, widget, member):
        """Sets a config attribute to the widget's active value"""
        setattr(self.config, member, widget.get_active())

    def _as_enabled_toggled(self, checkbox, *widgets):
        if checkbox.get_active():
            self.scrobbler.import_module(True)
        if self.scrobbler.imported():
            self.config.as_enabled = checkbox.get_active()
            self.scrobbler.init()
            for widget in widgets:
                widget.set_sensitive(self.config.as_enabled)
        elif checkbox.get_active():
            checkbox.set_active(False)

    def _as_username_changed(self, entry):
        if self.scrobbler.imported():
            self.config.as_username = entry.get_text()
            self.scrobbler.auth_changed()

    def _as_password_changed(self, entry):
        if self.scrobbler.imported():
            self.config.as_password_md5 = hashlib.md5(
                entry.get_text().encode('utf-8')).hexdigest()
            self.scrobbler.auth_changed()

    def _nameentry_changed(self, entry, profile_combo, remove_profiles):
        if not self.updating_nameentry:
            profile_num = profile_combo.get_active()
            self.config.profile_names[profile_num] = entry.get_text()
            self._populate_profile_combo(profile_combo, profile_num, remove_profiles)

    def _hostentry_changed(self, entry, profile_combo):
        profile_num = profile_combo.get_active()
        self.config.host[profile_num] = entry.get_text()

    def _portentry_changed(self, entry, profile_combo):
        profile_num = profile_combo.get_active()
        self.config.port[profile_num] = entry.get_value_as_int()

    def _passwordentry_changed(self, entry, profile_combo):
        profile_num = profile_combo.get_active()
        self.config.password[profile_num] = entry.get_text()

    def _direntry_changed(self, entry, profile_combo):
        profile_num = profile_combo.get_active()
        self.config.musicdir[profile_num] = misc.sanitize_musicdir(entry.get_filename())

    def _add_profile(self, _button, nameentry, profile_combo, remove_profiles):
        self.updating_nameentry = True
        profile_num = profile_combo.get_active()
        self.config.profile_names.append(_("New Profile"))
        nameentry.set_text(self.config.profile_names[-1])
        self.updating_nameentry = False
        self.config.host.append(self.config.host[profile_num])
        self.config.port.append(self.config.port[profile_num])
        self.config.password.append(self.config.password[profile_num])
        self.config.musicdir.append(self.config.musicdir[profile_num])
        self._populate_profile_combo(profile_combo, len(self.config.profile_names)-1, remove_profiles)
        self.populate_profiles_for_menu()

    def _remove_profile(self, _button, profile_combo, remove_profiles):
        profile_num = profile_combo.get_active()
        if profile_num == self.config.profile_num:
            # Profile deleted, revert to first profile:
            self.config.profile_num = 0
            self.reconnect(None)
        self.config.profile_names.pop(profile_num)
        self.config.host.pop(profile_num)
        self.config.port.pop(profile_num)
        self.config.password.pop(profile_num)
        self.config.musicdir.pop(profile_num)
        self.populate_profiles_for_menu()
        if profile_num > 0:
            self._populate_profile_combo(profile_combo, profile_num-1, remove_profiles)
        else:
            self._populate_profile_combo(profile_combo, 0, remove_profiles)

    def _profile_chosen(self, profile_combo, nameentry, hostentry, portentry, passwordentry, direntry):
        profile_num = profile_combo.get_active()
        self.updating_nameentry = True
        nameentry.set_text(str(self.config.profile_names[profile_num]))
        self.updating_nameentry = False
        hostentry.set_text(str(self.config.host[profile_num]))
        portentry.set_value(self.config.port[profile_num])
        passwordentry.set_text(str(self.config.password[profile_num]))
        direntry.set_current_folder(misc.sanitize_musicdir(self.config.musicdir[profile_num]))

    def _populate_profile_combo(self, profile_combo, active_index, remove_profiles):
        new_model = Gtk.ListStore(str)
        new_model.clear()
        profile_combo.set_model(new_model)
        for i, profile_name in enumerate(self.config.profile_names):
            combo_text = "[%s] %s" % (i+1, profile_name[:15])
            if len(profile_name) > 15:
                combo_text += "..."
            profile_combo.append_text(combo_text)
        profile_combo.set_active(active_index)
        # Enable remove button if there is more than one profile
        remove_profiles.set_sensitive(len(self.config.profile_names) > 1)

    def _lyrics_location_changed(self, combobox):
        self.config.lyrics_location = combobox.get_active()

    def _art_location_changed(self, combobox):
        if combobox.get_active() == consts.ART_LOCATION_CUSTOM:
            dialog = self.builder.get_object('custom_art_dialog')
            # Prompt user for playlist name:
            entry = self.builder.get_object('custom_art_entry')
            dialog.vbox.show_all()
            response = dialog.run()
            if response == Gtk.ResponseType.ACCEPT:
                self.config.art_location_custom_filename = entry.get_text().replace("/", "")
                iter = combobox.get_active_iter()
                model = combobox.get_model()
                model.set(iter, 0, "SONG_DIR/" + (self.config.art_location_custom_filename or _("custom")))
            else:
                # Revert to non-custom item in combobox:
                combobox.set_active(self.config.art_location)
            dialog.hide()
        self.config.art_location = combobox.get_active()

    def _crossfadecheck_toggled(self, button, *widgets):
        button_active = button.get_active()
        for widget in widgets:
            widget.set_sensitive(button_active)

    def _notiflocation_changed(self, combobox):
        self.config.traytips_notifications_location = combobox.get_active()
        self.renotify()

    def _notiftime_changed(self, combobox):
        self.config.popup_option = combobox.get_active()
        self.renotify()

    def _infofile_toggled(self, button, infofileformatbox):
        self.config.use_infofile = button.get_active()
        infofileformatbox.set_sensitive(self.config.use_infofile)
        if self.config.use_infofile:
            self.reinfofile()

    def plugin_click(self, _widget, event):
        if event.button == 3:
            self.plugin_UIManager.get_widget('/pluginmenu').popup(None, None, None, None, event.button, event.time)

    def plugin_toggled(self, _renderer, path, user_data):
        model, column = user_data
        enabled = not model[path][column]
        plugin = pluginsystem.get_info()[int(path)]
        pluginsystem.set_enabled(plugin, enabled)

        if enabled:
            # test that the plugin loads or already was loaded
            if not plugin.force_loaded():
                enabled = False
                pluginsystem.set_enabled(plugin, enabled)

        model[path][column] = enabled

    def plugin_about(self, _widget):
        plugin = self.plugin_get_selected()
        iconpb = self.plugin_get_icon_pixbuf(plugin)

        about_text = plugin.longname + "\n" + plugin.author + "\n"
        if len(plugin.author_email) > 0:
            about_text += "<" + plugin.author_email + ">"

        self.about_dialog = Gtk.AboutDialog()
        self.about_dialog.set_name(plugin.longname)
        self.about_dialog.set_role('about')
        self.about_dialog.set_version(plugin.version_string)
        if len(plugin.description.strip()) > 0:
            self.about_dialog.set_comments(plugin.description)
        if len(plugin.author.strip()) > 0:
            author = plugin.author
            if len(plugin.author.strip()) > 0:
                author += ' <' + plugin.author_email + '>'
            self.about_dialog.set_authors([author])
        if len(plugin.url.strip()) > 0:
            self.about_dialog.connect("activate-link", self.plugin_show_website)
            self.about_dialog.set_website(plugin.url)
        self.about_dialog.set_logo(iconpb)

        self.about_dialog.connect('response', self.plugin_about_close)
        self.about_dialog.connect('delete_event', self.plugin_about_close)
        self.about_dialog.show_all()

    def plugin_about_close(self, _event, _data=None):
        self.about_dialog.hide()
        return True

    def plugin_show_website(self, _dialog, link):
        return misc.browser_load(link, self.config.url_browser, \
                                 self.window)

    def plugin_configure(self, _widget):
        plugin = self.plugin_get_selected()
        ui.show_msg(self.prefswindow, "Nothing yet implemented.", "Configure", "pluginConfigure", Gtk.ButtonsType.CLOSE)

    def plugin_get_selected(self):
        model, i = self.pluginselection.get_selected()
        plugin_num = model.get_path(i).get_indices()[0]
        return pluginsystem.get_info()[plugin_num]

    def plugin_get_icon_pixbuf(self, plugin):
        pb = plugin.iconurl
        try:
            pb = GdkPixbuf.Pixbuf.new_from_file(pb)
        except:
            pb = self.pluginview.render_icon(Gtk.STOCK_EXECUTE, Gtk.IconSize.LARGE_TOOLBAR)
        return pb



# Copyright 2006-2009 Scott Horowitz <stonecrest@gmail.com>
# Copyright 2009-2014 Jonathan Ballet <jon@multani.info>
#
# This file is part of Sonata.
#
# Sonata is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Sonata is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Sonata.  If not, see <http://www.gnu.org/licenses/>.

"""Handle the mpd current playlist and provides a user interface for it."""

import os
import re
import urllib.parse, urllib.request

from gi.repository import Gtk, Gdk, Pango, GLib

from sonata import ui, misc, formatting
from sonata.mpdhelper import MPDSong


class Current:
    def __init__(self, config, mpd, name, on_button_press,
                 connected, sonata_loaded, songinfo, update_statusbar,
                 iterate_now, add_tab):
        self.config = config
        self.mpd = mpd
        self.on_button_press = on_button_press
        self.connected = connected
        self.sonata_loaded = sonata_loaded
        self.songinfo = songinfo
        self.update_statusbar = update_statusbar
        self.iterate_now = iterate_now

        self.store = None
        self.filterbox_visible = False
        self.update_skip = False
        self.columnformat = None
        self.columns = None

        self.refilter_handler_id = None
        # TreeViewColumn, order
        self.column_sorted = (None, Gtk.SortType.DESCENDING)
        self.total_time = 0
        self.resizing_columns = None
        self.prev_boldrow = -1
        self.playlist_pos_before_filter = None
        self.sel_rows = None

        # Current tab
        builder = ui.builder('current')
        self.view = builder.get_object('current_page_treeview')
        self.selection = self.view.get_selection()
        self.filterpattern = builder.get_object('current_page_filterbox_entry')
        self.filterbox = builder.get_object('current_page_filterbox')
        self.vbox = builder.get_object('current_page_v_box')
        builder.get_object('current_tab_label').set_text(name)

        tab_label_widget = builder.get_object('current_tab_eventbox')
        self.tab = add_tab(self.vbox, tab_label_widget, name, self.view)

        self.view.connect('drag-data-received', self.on_dnd_received)
        self.view.connect('row-activated', self.on_click)
        self.view.connect('button-press-event', self.on_button_press)
        self.view.connect('drag-begin', self.on_drag_begin)
        self.view.connect_after('drag-begin', self.on_dnd_after_drag_begin)
        self.view.connect('button-release-event', self.on_button_release)

        self.filter_changed_handler = self.filterpattern.connect(
            'changed', self.searchfilter_key_pressed)
        self.filterpattern.connect('activate', self.searchfilter_on_enter)
        filterclosebutton = builder.get_object(
            'current_page_filterbox_closebutton')
        filterclosebutton.connect('clicked', self.searchfilter_toggle)

        # Set up current view
        self.initialize_columns()
        self.selection.set_mode(Gtk.SelectionMode.MULTIPLE)

        target_reorder = ('MY_TREE_MODEL_ROW', Gtk.TargetFlags.SAME_WIDGET, 0)
        target_file_managers = ('text/uri-list', 0, 0)

        self.view.enable_model_drag_source(
            Gdk.ModifierType.BUTTON1_MASK,
            [target_reorder, target_file_managers],
            Gdk.DragAction.COPY | Gdk.DragAction.DEFAULT)
        self.view.enable_model_drag_dest(
            [target_reorder, target_file_managers],
            Gdk.DragAction.MOVE | Gdk.DragAction.DEFAULT)

        self.view.connect('drag-data-get', self.dnd_get_data_for_file_managers)

    def get_widgets(self):
        return self.vbox

    def is_empty(self):
        return len(self.store) == 0

    def clear(self):
        self.store.clear()

    def on_song_change(self, status):
        self.unbold_boldrow(self.prev_boldrow)

        if status and 'song' in status:
            row = int(status['song'])
            self.boldrow(row)
            self.center_song_in_list()
            self.prev_boldrow = row

    def try_keep_position(func):
        """Decorator to keep the position of the view while updating it"""

        def do_try_keep_position(self, *args, **kwargs):
            realized = self.view.get_realized()
            if realized:
                position = self.view.get_visible_rect()

            result = func(self, *args, **kwargs)

            if realized:
                self.view.scroll_to_point(-1, position.y)

            return result
        return do_try_keep_position

    def get_treeview(self):
        return self.view

    def get_selection(self):
        return self.selection

    def get_filterbox_visible(self):
        return self.filterbox_visible

    @try_keep_position
    def initialize_columns(self):
        # Initialize current playlist data and widget
        self.resizing_columns = False
        self.columnformat = self.config.currentformat.split("|")
        current_columns = [MPDSong] + [str] * len(self.columnformat) + [int]
        previous_tracks = (item[0] for item in (self.store or []))
        self.store = Gtk.ListStore(*(current_columns))
        cellrenderer = Gtk.CellRendererText()
        cellrenderer.set_property("ellipsize", Pango.EllipsizeMode.END)
        cellrenderer.set_property("weight-set", True)

        num_columns = len(self.columnformat)
        if num_columns != len(self.config.columnwidths):
            # Number of columns changed, set columns equally spaced:
            self.config.columnwidths = [self.view.get_allocation().width / \
                                        num_columns] * num_columns

        colnames = formatting.parse_colnames(self.config.currentformat)
        for column in self.view.get_columns():
            self.view.remove_column(column)
        self.columns = [Gtk.TreeViewColumn(name, cellrenderer, markup=(i + 1))
                for i, name in enumerate(colnames)]
        for tree in self.columns:
            tree.add_attribute(cellrenderer, "weight", len(current_columns) - 1)

        for column, width in zip(self.columns, self.config.columnwidths):
            column.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
            # If just one column, we want it to expand with the tree, so
            # don't set a fixed_width; if multiple columns, size accordingly:
            if num_columns > 1:
                column.set_resizable(True)
                try:
                    column.set_fixed_width(max(width, 10))
                except:
                    column.set_fixed_width(150)
            column.connect('clicked', self.on_column_click)
            self.view.append_column(column)

        self.view.set_fixed_height_mode(True)
        self.view.set_headers_visible(num_columns > 1 and \
                                         self.config.show_header)
        self.view.set_headers_clickable(not self.filterbox_visible)
        self.update_format(previous_tracks)
        self.view.set_model(self.store)

    def dnd_get_data_for_file_managers(self, _treeview, context, selection,
                                       _info, timestamp):

        if not os.path.isdir(self.config.current_musicdir):
            # Prevent the DND mouse cursor from looking like we can DND
            # when we clearly can't.
            return

        Gdk.drag_status(context, Gdk.DragAction.COPY, timestamp)

        filenames = self.get_selected_filenames(True)
        uris = ["file://%s" % urllib.parse.quote(filename)
            for filename in filenames]

        selection.set_uris(uris)

    def get_selected_filenames(self, return_abs_paths):
        _model, selected = self.selection.get_selected_rows()
        filenames = []

        for path in selected:
            index = path.get_indices()[0]
            item = self.store[index][0].file
            if return_abs_paths:
                filenames.append(
                    os.path.join(self.config.current_musicdir, item))
            else:
                filenames.append(item)
        return filenames

    def update_format(self, tracks):
        for i, track in enumerate(tracks):
            items = [formatting.parse(part, track, True)
                     for part in self.columnformat]

            if self.songinfo().pos == i:
                weight = [Pango.Weight.BOLD]
            else:
                weight = [Pango.Weight.NORMAL]

            self.store.append([track] + items + weight)

    @try_keep_position
    def current_update(self, prevstatus_playlist, new_playlist_length):
        if self.connected():
            self.view.freeze_child_notify()
            self.unbold_boldrow(self.prev_boldrow)

            if not self.update_skip:
                save_model = self.view.get_model()
                self.view.set_model(None)
                if prevstatus_playlist:
                    changed_songs = self.mpd.plchanges(prevstatus_playlist)
                else:
                    changed_songs = self.mpd.plchanges(0)


                newlen = int(new_playlist_length)
                currlen = len(self.store)

                for track in changed_songs:
                    pos = track.pos

                    items = [formatting.parse(part, track, True)
                             for part in self.columnformat]

                    if pos < currlen:
                        # Update attributes for item:
                        i = self.store.get_iter((pos, ))
                        if track.id != self.store.get_value(i, 0).id:
                            self.store.set_value(i, 0, track)
                        for index in range(len(items)):
                            if items[index] != self.store.get_value(i, index+1):
                                self.store.set_value(i, index + 1, items[index])
                    else:
                        # Add new item:
                        self.store.append(
                            [track] + items + [Pango.Weight.NORMAL])

                if newlen == 0:
                    self.store.clear()
                else:
                    # Remove excess songs:
                    for i in range(currlen - newlen):
                        it = self.store.get_iter((currlen - 1 - i,))
                        self.store.remove(it)

                self.view.set_model(save_model)
            self.update_skip = False

            # Update statusbar time:
            self.total_time = sum(item[0].time for item in self.store)

            if 'pos' in self.songinfo():
                currsong = self.songinfo().pos
                self.boldrow(currsong)
                self.prev_boldrow = currsong

            self.view.thaw_child_notify()
            self.header_update_column_indicators()
            self.update_statusbar()
            ui.change_cursor(None)

    def header_update_column_indicators(self):
        # If we just sorted a column, display the sorting arrow:
        if self.column_sorted[0]:
            if self.column_sorted[1] == Gtk.SortType.DESCENDING:
                self.header_hide_all_indicators(self.view, True)
                self.column_sorted[0].set_sort_order(Gtk.SortType.ASCENDING)
                self.column_sorted = (None, Gtk.SortType.ASCENDING)
            else:
                self.header_hide_all_indicators(self.view, True)
                self.column_sorted[0].set_sort_order(Gtk.SortType.DESCENDING)
                self.column_sorted = (None, Gtk.SortType.DESCENDING)

    def header_hide_all_indicators(self, treeview, show_sorted_column):
        if not show_sorted_column:
            self.column_sorted = (None, Gtk.SortType.descending)
        for column in treeview.get_columns():
            if show_sorted_column and column == self.column_sorted[0]:
                column.set_sort_indicator(True)
            else:
                column.set_sort_indicator(False)

    def center_song_in_list(self, do_select_current_song=False):
        if not self.filterbox_visible and self.config.expanded and \
           len(self.store) > 0:
            row_path = Gtk.TreePath(self.songinfo().pos)
            if do_select_current_song:
                selection = self.view.get_selection()
                selection.unselect_all()
                selection.select_path(row_path)
                self.view.set_cursor(row_path)
                self.view.grab_focus()

            # For some reason, calling view.set_cursor() after
            # view.scroll_to_cell() changes the scrolling position of the view
            # (the selected item appears at the top of the view, not in the
            # center anymore).
            self.view.scroll_to_cell(row_path, None, True, 0.5, 0.5)

    def get_songid(self, i, model):
        return model.get_value(i, 0).id

    def on_drag_begin(self, _widget, _context):
        self.sel_rows = False

    def on_dnd_after_drag_begin(self, _widget, context):
        # Override default image of selected row with sonata icon:
        Gtk.drag_set_icon_stock(context, 'sonata', 0, 0)

    def on_button_release(self, widget, event):
        if self.sel_rows:
            self.sel_rows = False
            # User released mouse, select single row:
            selection = widget.get_selection()
            selection.unselect_all()
            path, _col, _x, _y = widget.get_path_at_pos(int(event.x),
                                                        int(event.y))
            selection.select_path(path)

    def on_column_click(self, column):
        columns = self.view.get_columns()
        col_num = 0
        for col in columns:
            col_num = col_num + 1
            if column == col:
                self.sort('col' + str(col_num), column)
                return

    def on_sort_by_artist(self, _action):
        self.sort('artist')

    def on_sort_by_album(self, _action):
        self.sort('album')

    def on_sort_by_title(self, _action):
        self.sort('title')

    def on_sort_by_file(self, _action):
        self.sort('file')

    def on_sort_by_dirfile(self, _action):
        self.sort('dirfile')

    def sort(self, mode, column=None):
        if self.connected():
            if not self.store:
                return

            while Gtk.events_pending():
                Gtk.main_iteration()
            songs = []
            track_num = 0

            if mode[0:3] == 'col':
                col_num = int(mode.replace('col', ''))
                if column.get_sort_indicator():
                    # If this column was already sorted, reverse list:
                    self.column_sorted = (column, self.column_sorted[1])
                    self.on_sort_reverse(None)
                    return
                else:
                    self.column_sorted = (column, Gtk.SortType.DESCENDING)
                mode = "col"

            # If the first tag in the format is song length, we will make
            # sure to compare the same number of items in the song length
            # string (e.g. always use ##:##:##) and pad the first item to two
            # (e.g. #:##:## -> ##:##:##)
            custom_sort = False
            if mode == 'col':
                custom_sort, custom_pos = self.sort_get_first_format_tag(
                    self.config.currentformat, col_num, 'L')

            for track in (item[0] for item in self.store):
                record = {}
                # Those items that don't have the specified tag will be put at
                # the end of the list (hence the 'zzzzzzz'):
                zzz = 'zzzzzzzz'
                if mode == 'artist':
                    record["sortby"] = (
                        misc.lower_no_the(track.artist or zzz),
                        (track.album or zzz).lower(),
                        track.disc,
                        track.track)
                elif mode == 'album':
                    record["sortby"] = ((track.album or zzz).lower(),
                                        track.disc,
                                        track.track)
                elif mode == 'file':
                    record["sortby"] = os.path.basename(track.file or
                                                        zzz).lower()
                elif mode == 'dirfile':
                    record["sortby"] = (track.file or zzz).lower()
                elif mode == 'col':
                    # Sort by column:
                    record["sortby"] = self.store.get_value(
                        self.store.get_iter((track_num, 0)),
                        col_num).lower()
                    if custom_sort:
                        record["sortby"] = self.sanitize_songlen_for_sorting(
                            record["sortby"], custom_pos)
                else:
                    record["sortby"] = track.get(mode, zzz).lower()

                record["id"] = track.id
                songs.append(record)
                track_num = track_num + 1

            songs.sort(key=lambda x: x["sortby"])

            pos = 0
            self.mpd.command_list_ok_begin()
            for item in songs:
                self.mpd.moveid(item["id"], pos)
                pos += 1
            self.mpd.command_list_end()
            self.iterate_now()

            self.header_update_column_indicators()

    def sort_get_first_format_tag(self, format, colnum, tag_letter):
        # Returns a tuple with whether the first tag of the format
        # includes tag_letter and the position of the tag in the string:
        formats = format.split('|')
        format = formats[colnum-1]
        prev_letter = None
        for letter in format:
            if letter == tag_letter and prev_letter == '%':
                return (True, pos)
            else:
                break
            prev_letter = letter
        return (False, 0)

    def sanitize_songlen_for_sorting(self, songlength, pos_of_string):
        songlength = songlength[pos_of_string:]
        items = [item.zfill(2) for item in songlength.split(':')]
        for i in range(3 - len(items)):
            items.insert(0, "00")
        return ":".join(item for item in items[:3])

    def on_sort_reverse(self, _action):
        if self.connected():
            if not self.store:
                return
            while Gtk.events_pending():
                Gtk.main_iteration()
            top = 0
            bot = len(self.store)-1
            self.mpd.command_list_ok_begin()
            while top < bot:
                self.mpd.swap(top, bot)
                top = top + 1
                bot = bot - 1
            self.mpd.command_list_end()
            self.iterate_now()

    def on_dnd_received(self, treeview, drag_context, x, y, selection, _info, timestamp):
        drop_info = treeview.get_dest_row_at_pos(x, y)

        if selection.get_data():
            if not os.path.isdir(self.config.current_musicdir):
                return
            # DND from outside sonata:
            uri = selection.get_data().strip().decode('utf-8')
            path = urllib.request.url2pathname(uri)
            paths = path.rsplit('\n')
            mpdpaths = []
            # Strip off paranthesis so that we can DND entire music dir
            # if we wish.
            musicdir = self.config.current_musicdir[:-1]
            for i, path in enumerate(paths):
                paths[i] = path.rstrip('\r')
                if paths[i].startswith('file://'):
                    paths[i] = paths[i][7:]
                elif paths[i].startswith('file:'):
                    paths[i] = paths[i][5:]
                if paths[i].startswith(musicdir):
                    paths[i] = paths[i][len(musicdir):]
                    if len(paths[i]) == 0:
                        paths[i] = "/"
                    listallinfo = self.mpd.listallinfo(paths[i])
                    for item in listallinfo:
                        if 'file' in item:
                            mpdpaths.append(item['file'])

                # Add local file, available in mpd 0.14. This currently
                # work because python-mpd does not support unix socket
                # paths, won't which is needed for authentication for
                # local files. It's also therefore untested.
                if os.path.isdir(paths[i]):
                    filenames = misc.get_files_recursively(paths[i])
                else:
                    filenames = [paths[i]]
                for filename in filenames:
                    if os.path.exists(filename):
                        mpdpaths.append("file://" + urllib.parse.quote(filename))
            if len(mpdpaths) > 0:
                # Items found, add to list at drop position:
                if drop_info:
                    destpath, position = drop_info
                    if position in (Gtk.TreeViewDropPosition.BEFORE,
                                    Gtk.TreeViewDropPosition.INTO_OR_BEFORE):
                        songid = destpath[0]
                    else:
                        songid = destpath[0] + 1
                else:
                    songid = len(self.store)
                for mpdpath in mpdpaths:
                    self.mpd.addid(mpdpath, songid)
            self.iterate_now()
            return

        # Otherwise, it's a DND just within the current playlist
        model = self.store
        _foobar, selected = self.selection.get_selected_rows()

        # calculate all this now before we start moving stuff
        drag_sources = []
        for path in selected:
            index = path[0]
            treeiter = model.get_iter(path)
            songid = self.get_songid(treeiter, model)
            drag_sources.append([index, treeiter, songid])

        # Keep track of the moved iters so we can select them afterwards
        moved_iters = []

        # Will manipulate model to prevent the entire playlist from refreshing
        offset = 0
        self.mpd.command_list_ok_begin()
        for index, treeiter, songid in drag_sources:
            if drop_info:
                destpath, position = drop_info
                dest = destpath[0] + offset
                if dest < index:
                    offset = offset + 1
                pop_from = index
                move_to = dest
                if position in (Gtk.TreeViewDropPosition.BEFORE,
                                Gtk.TreeViewDropPosition.INTO_OR_BEFORE):
                    insert_to = dest
                    if dest < index + 1:
                        pop_from = index + 1
                    else:
                        move_to = dest - 1
                else:
                    insert_to = dest + 1
                    if dest < index:
                        pop_from = index + 1
                        move_to = insert_to
            else:
                dest = len(self.store) - 1
                insert_to = dest + 1

            self.mpd.moveid(songid, move_to)
            moved_iters.append(model.insert(insert_to, tuple(model[index])))
            model.remove(treeiter)

            # now fixup
            for source in drag_sources:
                if dest < index:
                    # we moved it back, so all indexes inbetween increased by 1
                    if dest < source[0] < index:
                        source[0] += 1
                else:
                    # we moved it ahead, so all indexes inbetween
                    # decreased by 1
                    if index < source[0] < dest:
                        source[0] -= 1
        self.mpd.command_list_end()

        # we are manipulating the model manually for speed, so...
        self.update_skip = True

        # Gdk.DragContext.get_action() returns a bitmask of actions
        if drag_context.get_actions() & Gdk.DragAction.MOVE:
            Gdk.drag_finish(drag_context, True, True, timestamp)
            self.header_hide_all_indicators(self.view, False)
        self.iterate_now()

        selection = treeview.get_selection()
        selection.unselect_all()
        for i in moved_iters:
            selection.select_iter(i)

        if moved_iters:
            treeview.scroll_to_cell(model.get_path(moved_iters[0]), None)

    def on_click(self, _treeview, path, _column):
        model = self.view.get_model()
        if self.filterbox_visible:
            self.searchfilter_on_enter(None)
            return
        try:
            i = model.get_iter(path)
            self.mpd.playid(self.get_songid(i, model))
        except:
            pass
        self.sel_rows = False
        self.iterate_now()

    def searchfilter_toggle(self, _widget, initial_text=""):
        if self.filterbox_visible:
            ui.hide(self.filterbox)
            self.filterbox_visible = False
            self.filterpattern.set_text("")
            self.view.set_model(self.store)
        elif self.connected():
            self.playlist_pos_before_filter = \
                    self.view.get_visible_rect().height
            self.filterbox_visible = True
            with self.filterpattern.handler_block(self.filter_changed_handler):
                self.filterpattern.set_text(initial_text)
            ui.show(self.filterbox)
            self.filterpattern.grab_focus()
        self.view.set_headers_clickable(not self.filterbox_visible)

    def model_filter_func(self, model, iter, regex):
        row = model.get(iter, 1, *range(len(self.columnformat)- 1)[1:])
        for cell in row:
            if regex.match(cell.lower()):
                return True
        return False

    def searchfilter_on_enter(self, _entry):
        model, selected = self.view.get_selection().get_selected_rows()
        song_id = None
        if len(selected) > 0:
            # If items are selected, play the first selected item:
            song_id = self.get_songid(model.get_iter(selected[0]), model)
        elif len(model) > 0:
            # If nothing is selected: play the first item:
            song_id = self.get_songid(model.get_iter_first(), model)
        if song_id:
            self.searchfilter_toggle(None)
            self.mpd.playid(song_id)

    def searchfilter_key_pressed(self, widget):
        # We have something new to search, try first to cancel the previous
        # search.
        try:
            GLib.source_remove(self.refilter_handler_id)
        except TypeError: # self.refilter_handler_id is None
            pass

        text = widget.get_text()
        if text == '':
            # Nothing to search for, just display the whole model.
            self.view.set_model(self.store)
            return

        regex = misc.escape_html(text)
        regex = re.escape(regex)
        regex = '.*' + regex.replace(' ', ' .*').lower()
        filter_regex = re.compile(regex)

        def set_filtering_function(regex):
            # Creates a Gtk.TreeModelFilter
            filter_model = self.store.filter_new()
            filter_model.set_visible_func(self.model_filter_func, regex)
            self.view.set_model(filter_model)

        # Delay slightly the new search, in case something else is coming.
        self.refilter_handler_id = GLib.timeout_add(
            250, set_filtering_function, filter_regex)

    def filter_key_pressed(self, widget, event, treeview):
        if event.keyval == Gdk.keyval_from_name('Down') or \
           event.keyval == Gdk.keyval_from_name('Up') or \
           event.keyval == Gdk.keyval_from_name('Page_Down') or \
           event.keyval == Gdk.keyval_from_name('Page_Up'):

            treeview.grab_focus()
            treeview.emit("key-press-event", event)
            GLib.idle_add(self.filter_entry_grab_focus, widget)

    def boldrow(self, row):
        if row > -1:
            try:
                self.store[row][-1] = Pango.Weight.BOLD
            except IndexError:
                # The row might not exist anymore
                pass

    def unbold_boldrow(self, row):
        if row > -1:
            try:
                self.store[row][-1] = Pango.Weight.NORMAL
            except IndexError:
                # The row might not exist anymore
                pass

    def on_remove(self):
        model, selected = self.selection.get_selected_rows()
        if len(selected) == len(self.store) and not self.filterbox_visible:
            # Everything is selected, clear:
            self.mpd.clear()
        elif len(selected) > 0:
            # we are manipulating the model manually for speed, so...
            self.update_skip = True
            self.mpd.command_list_ok_begin()
            for i in (model.get_iter(path) for path in reversed(selected)):
                self.mpd.deleteid(model.get(i, 0)[0].id)
                if model != self.store:
                    # model is different if there is a filter currently applied.
                    # So we retrieve the iter of the wrapped model...
                    i = model.convert_iter_to_child_iter(i)
                self.store.remove(i)
            self.mpd.command_list_end()
